/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.binding;

import gnu.trove.TObjectIntHashMap;

import java.util.Collection;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.Set;

import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.RuntimeBindingException;
import org.simantics.databoard.binding.java.ArrayListBinding;
import org.simantics.databoard.binding.java.ByteArrayBinding;
import org.simantics.databoard.binding.java.DoubleArrayBinding;
import org.simantics.databoard.binding.java.FloatArrayBinding;
import org.simantics.databoard.binding.java.IntArrayBinding;
import org.simantics.databoard.binding.java.LinkedListBinding;
import org.simantics.databoard.binding.java.LongArrayBinding;
import org.simantics.databoard.binding.java.ObjectArrayBinding;
import org.simantics.databoard.binding.mutable.MutableInteger;
import org.simantics.databoard.type.ArrayType;
import org.simantics.databoard.util.IdentityPair;
import org.simantics.databoard.util.Range;

/**
 * This is a binding of Array type and a Java Object.
 * 
 * @see ArrayType
 * @see ArrayListBinding for Binding of java.util.ArrayList<?>
 * @see LinkedListBinding for Binding of java.util.LinkedList<?>
 * @see ObjectArrayBinding for Binding of Object[]
 * @see IntArrayBinding for primitive array int[] binding
 * @see ByteArrayBinding for primitive array byte[] binding
 * @see LongArrayBinding for primitive array long[] binding
 * @see DoubleArrayBinding for primitive array double[] binding
 * @see FloatArrayBinding for primitive array float[] binding 
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public abstract class ArrayBinding extends Binding {

    protected Binding componentBinding;
    
	public ArrayBinding(Binding componentBinding) {
        this.componentBinding = componentBinding;
        this.type = new ArrayType(componentBinding.type);
    }
	
	public ArrayBinding(ArrayType type, Binding componentBinding) {
		if (!type.getComponentType().equals(componentBinding.type()))
			throw new IllegalArgumentException("Binding for "+type.componentType+" expected, got "+componentBinding.type());
        this.componentBinding = componentBinding;
        this.type = type;
    }

	/**
	 * Returns true if array length can be modified. 
	 * 
	 * @return true if array length can be modified, false if not
	 */
	@Override
	public boolean isImmutable() {
		return super.isImmutable();
	}
	
	@Override
	public ArrayType type() {
		return (ArrayType) type;
	}

    public Binding getComponentBinding() {
        return componentBinding;
    }
    
    /**
     * Create a new empty array
     * @return array object
     */
    public abstract Object create();
	
    /**
     * Create a new array with initial values copied or referred from a collection. 
     * 
     * @param collection
     * @return array object
     * @throws BindingException
     */
	public Object create(Collection<Object> collection) throws BindingException {
		return create(collection.size(), collection.iterator());
	}
    
    /**
     * Create new array instance with initial values possibly borrowed from an interator.
     * <p>
     * The implementation iterate the iterator before returning.
     * 
     * @param length array length
     * @param values value iterator 
     * @return new instance
     */
	public abstract Object create(int length, Iterator<Object> values)
	throws BindingException;
		
	/**
	 * Create Array with initial values possibly borrowed from an java.lang.Array
	 * 
	 * @param array
	 * @return array of ArrayType
	 * @throws BindingException
	 */
	public abstract Object create(Object...array)
	throws BindingException;
	
	public Object createUnchecked(Object...array)
	throws RuntimeBindingException {
		try {
			return create(array);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
	public void add(Object array, Object element) throws BindingException 
	{
		int size = size(array);
		add(array, size, element);
	}
	public abstract void add(Object array, int index, Object element) throws BindingException, IndexOutOfBoundsException;
	public void remove(Object array, int index) throws BindingException, IndexOutOfBoundsException
	{
		remove(array, index, 1);
	}
	public abstract void remove(Object array, int index, int count) throws BindingException, IndexOutOfBoundsException;
	public abstract Object get(Object array, int index) throws BindingException, IndexOutOfBoundsException;		
	public abstract void getAll(Object array, Object[] result) throws BindingException;
	public abstract void set(Object array, int index, Object value) throws BindingException;
	
	public abstract int size(Object array) throws BindingException;
	
	@Override
	public void accept(Visitor1 v, Object obj) {
	    v.visit(this, obj);        
	}

	@Override
	public <T> T accept(Visitor<T> v) {
	    return v.visit(this);
	}
	
    /**
     * Assert the instance is valid and follows restrictions set in data type.
     * Assertions:
     *   1. The Length of the array meets ArrayType range
     *   2. Recursive assertion of each element
     * 
     * @param obj the instance
     * @throws BindingException on invalid instance
     */
    public void assertInstaceIsValid(Object obj, Set<Object> validInstances)
    throws BindingException 
    {
    	ArrayType type = type();
    	// Assert The Length of the array meets ArrayType range
    	int length = size(obj);
    	Range range = type.getLength();
    	if (range!=null && !range.contains(length)) {
    		throw new BindingException("Array length (="+length+")out of bounds. "+range+" was expected.");
    	}
    	// Recursive assertion of each element
    	for (int i=0; i<length; i++)
    	{
    		Object component = get(obj, i);
    		componentBinding.assertInstaceIsValid(component, validInstances);
    	}
    }
    
    @Override
    public int deepHashValue(Object value, IdentityHashMap<Object, Object> hashedObjects) throws BindingException {
    	int result = 1;
    	int len = size(value);
    	for (int i=0; i<len; i++) {
    		Object element = get(value, i);
    		result = 31*result + componentBinding.deepHashValue(element, hashedObjects);
    	}
    	return result;
    }
    
    @Override
    public int deepCompare(Object o1, Object o2,
    		Set<IdentityPair<Object, Object>> compareHistory)
    		throws BindingException {
		// Compare Lengths
		int l1 = size(o1);
		int l2 = size(o2);
		int dif = l1 - l2;
		if (dif!=0) return dif;
		// Compare elements
		Binding c = getComponentBinding();
		for (int i=0; i<l1; i++) {
			Object e1 = get(o1, i);
			Object e2 = get(o2, i);
			dif = c.deepCompare(e1, e2, compareHistory);
			if (dif!=0) return dif;
		}
		return 0;
    }

    @Override
    protected void toString(Object value, StringBuilder b, TObjectIntHashMap<Object> refs, MutableInteger nameCount) throws BindingException {
		Binding componentBinding = getComponentBinding();
		b.append('[');
		int size = size(value);
		for(int i=0;i<size;++i) {
			if(i>0)
				b.append(", ");
			componentBinding.toString(get(value, i), b, refs, nameCount);
		}
		b.append(']');
	}
}

