/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.reflection;

import java.util.Map.Entry;
import java.util.Set;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.UnionBinding;
import org.simantics.databoard.binding.VariantBinding;
import org.simantics.databoard.binding.error.BindingConstructionException;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.bindingscheme.DataTypeConstructionException;
import org.simantics.databoard.reflection.ReflectionBinding.ClassBinding;
import org.simantics.databoard.reflection.ReflectionBinding.ClassDataType;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.RecordType;
import org.simantics.databoard.type.UnionType;

/**
 * ObjectVariantBinding binds VarianType and Object.class.
 * The implementation is immutable as Object.class is not a container, it is
 * a value. <p>
 * 
 * ObjectVariantBinding understands many classes 
 *   Collections: ArrayList, List, LinkedList, int[], .., other arrays [], Map, TreeMap, HashMap, Set, HashSet, TreeSet, 
 *   Strings: String
 *   immutable primitives: Integer, Boolean, ...
 *   mutable primitives: MutableInteger, MutableBoolean, ...
 *   classes: Record-classes, Bean-Classes, Immutable-Classes 
 *   <code>null</code> is bound to a void.
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class ObjectVariantBinding  extends VariantBinding {
	
	@Deprecated
	public static ObjectVariantBinding INSTANCE = new ObjectVariantBinding(Bindings.reflectionBinding); 
	
	ReflectionBinding reflection;
	
	public ObjectVariantBinding(ReflectionBinding reflection) {
		this.reflection = reflection;
	}
	
	@Override
	public Object create(Binding binding, Object value) throws BindingException {
		if (binding instanceof VoidBinding) return null;
		try {
			Binding rb = binding;
			Object rv = value;
			DataType type = rb.type();
			
			
			Binding lb = null;
			synchronized(ReflectionBinding.class) {
				for (Entry<ClassBinding, Binding> e : reflection.bindingCache.entrySet())
				{
					if (binding == e.getValue()) {
						Class<?> clazz = e.getKey().clazz.clazz;
						if (clazz.isPrimitive()) continue;
						lb = binding;
						break;
					}
				}
				
				if (lb==null) {
					for (Entry<ClassDataType, DataType> e : reflection.dataTypeCache.entrySet())
						if (e.getValue().equals(type)) {
							Class<?> clazz = e.getKey().clazz;
							if (clazz.isPrimitive()) continue;
							lb = reflection.getBinding( clazz );
							break;
						}
				}
			}
			
			if (lb==null) throw new BindingException("I don't know how to bind "+type+" to java.lang.Object");			
			Object lv = Bindings.adapt(rv, rb, lb);
			
			return lv;
		} catch (BindingConstructionException e) {
			throw new BindingException(e);
		} catch (AdaptException e) {
			throw new BindingException(e);
		} 		
	}

	@Override
	public Binding getContentBinding(Object variant) throws BindingException {
		try {
			if (variant==null) return VoidBinding.VOID_BINDING;
			return reflection.getBinding(variant.getClass());
		} catch (BindingConstructionException e) {
			throw new BindingException(e);
		}
	}

	@Override
	public DataType getContentType(Object variant) throws BindingException {
		try {
			if (variant==null) return RecordType.VOID_TYPE;
			return reflection.getDataType(variant.getClass());
		} catch (DataTypeConstructionException e) {
			throw new BindingException(e);
		}
	}

	@Override
	public Object getContent(Object variant, Binding binding)
	throws BindingException {
		try {
			Binding binding1 = getContentBinding(variant);
			Object value1 = variant;
		
			Binding binding2 = binding;
			Object value2 = Bindings.adapt(value1, binding1, binding2);
		
			return value2;
		} catch (AdaptException e) {
			throw new BindingException(e);
		}
	}
	
	@Override
	public Object getContent(Object variant)
	throws BindingException {
		return variant;
	}	

	@Override
	public void setContent(Object variant, Binding binding, Object value)
	throws BindingException {
		try {
			Binding binding1 = getContentBinding(variant);
			DataType type1 = binding1.type();
			
			Binding binding2 = binding;
			DataType type2 = binding2.type();
			
			// New value cannot be assigned to Object from different data type
			if (!type1.equals(type2)) {
				throw new BindingException("Cannot assign "+value.getClass()+" to "+variant.getClass());
			}

			// Can only assing to union or record of previous type
			if (type1 instanceof RecordType) {
				RecordType recordType = (RecordType) type1;
				int count = recordType.count();
				RecordBinding rb1 = (RecordBinding) binding1;
				RecordBinding rb2 = (RecordBinding) binding2;
				Object resultValues[] = new Object[count];
				for (int i=0; i<count; i++) {
					Binding cb1 = rb1.componentBindings[i];
					Binding cb2 = rb2.componentBindings[i];
										
					Object componentValue2 = rb2.getComponent(value, i);
					Object componentValue1 = Bindings.adapt(componentValue2, cb2, cb1);
					resultValues[i] = componentValue1;
				}
				rb1.setComponents(variant, resultValues);
			} else if (type1 instanceof UnionType) {
				UnionType unionType = (UnionType) type1;
				UnionBinding ub1 = (UnionBinding) binding1;
				UnionBinding ub2 = (UnionBinding) binding2;
				int tag = ub2.getTag(value);
				Binding cb1 = ub1.getComponentBindings()[tag];
				Binding cb2 = ub2.getComponentBindings()[tag];
				
				Object componentValue2 = ub2.getValue(value);
				Object componentValue1 = Bindings.adapt(componentValue2, cb2, cb1);
				
				ub1.setValue(variant, tag, componentValue1);
			} else {
				throw new BindingException("Cannot assign "+value.getClass()+" to "+variant.getClass());
			}
			
		} catch (AdaptException e) {
			throw new BindingException(e);
		}
			
	}

	@Override
	public void assertInstaceIsValid(Object obj, Set<Object> validInstances) throws BindingException {
		getContentBinding(obj).assertInstaceIsValid(obj);
	}

	@Override
	public boolean isInstance(Object obj) {
		return true;
	}
	
	@Override
	public boolean isImmutable() {
		return true;
	}
	
}

