/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.reflection;

import java.lang.reflect.Array;
import java.util.Iterator;

import org.simantics.databoard.binding.ArrayBinding;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.UnsupportedOperationException;
import org.simantics.databoard.type.ArrayType;

/**
 * ReflectionArrayBinding binds ArrayType to T[], Object[], 
 * boolean[], byte[], int[], long[], float[], and double[]. 
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class ReflectionArrayBinding extends ArrayBinding {
	
	Class<?> clazz;
	Class<?> componentClass;

	ReflectionArrayBinding(Class<?> clazz, ArrayType type, Binding componentBinding) {
		super(componentBinding);
		this.clazz = clazz;
		this.type = type;		
		this.componentClass = clazz.getComponentType();		
	}

	@Override
	public Object create() {
		return Array.newInstance(componentClass, 0);
	}
	
	@Override
	public Object create(int length, Iterator<Object> values) {
		Object result = Array.newInstance(componentClass, length);
		if (result instanceof int[]) {
			for (int i=0; i<length; i++) 
				Array.setInt(result, i, (Integer)values.next());
		} else if (result instanceof boolean[]) {
			for (int i=0; i<length; i++) 
				Array.setBoolean(result, i, (Boolean)values.next());
		} else if (result instanceof byte[]) {
			for (int i=0; i<length; i++) 
				Array.setByte(result, i, (Byte)values.next());
		} else if (result instanceof long[]) {
			for (int i=0; i<length; i++) 
				Array.setLong(result, i, (Long)values.next());
		} else if (result instanceof float[]) {
			for (int i=0; i<length; i++) 
				Array.setFloat(result, i, (Float)values.next());
		} else if (result instanceof double[]) {
			for (int i=0; i<length; i++) 
				Array.setDouble(result, i, (Double)values.next());
		} else {
			for (int i=0; i<length; i++) 
				Array.set(result, i, values.next());
		}
		
		return result;
	}
	
	@Override
	public Object create(Object...array) throws BindingException {
		int length = array.length;
		Object result = Array.newInstance(componentClass, length);
		if (result instanceof int[]) {
			for (int i=0; i<length; i++) 
				Array.setInt(result, i, (Integer)array[i]);
		} else if (result instanceof boolean[]) {
			for (int i=0; i<length; i++) 
				Array.setBoolean(result, i, (Boolean)array[i]);
		} else if (result instanceof byte[]) {
			for (int i=0; i<length; i++) 
				Array.setByte(result, i, (Byte)array[i]);
		} else if (result instanceof long[]) {
			for (int i=0; i<length; i++) 
				Array.setLong(result, i, (Long)array[i]);
		} else if (result instanceof float[]) {
			for (int i=0; i<length; i++) 
				Array.setFloat(result, i, (Float)array[i]);
		} else if (result instanceof double[]) {
			for (int i=0; i<length; i++) 
				Array.setDouble(result, i, (Double)array[i]);
		} else {
			System.arraycopy(array, 0, result, 0, length);
		}
		
		return result;
	}
	
	@Override
	public Object get(Object array, int index) throws BindingException {
		if (!isInstance(array)) throw new BindingException("Unexpected class "+array.getClass().getSimpleName()+", "+clazz.getSimpleName()+" expected");
		return Array.get(array, index);
	}
	
	@Override
	public void getAll(Object array, Object[] result) throws BindingException {
		int length = Array.getLength(array);
		
		if (array instanceof int[]) {
			int[] a = (int[]) array;
			for (int i=0; i<length; i++) 
				result[i] = a[i];
		} else if (array instanceof boolean[]) {
			boolean[] a = (boolean[]) array;
			for (int i=0; i<length; i++) 
				result[i] = a[i];
		} else if (array instanceof byte[]) {
			byte[] a = (byte[]) array;
			for (int i=0; i<length; i++) 
				result[i] = a[i];
		} else if (array instanceof long[]) {
			long[] a = (long[]) array;
			for (int i=0; i<length; i++) 
				result[i] = a[i];
		} else if (array instanceof float[]) {
			float[] a = (float[]) array;
			for (int i=0; i<length; i++) 
				result[i] = a[i];
		} else if (array instanceof double[]) {
			double[] a = (double[]) array;
			for (int i=0; i<length; i++) 
				result[i] = a[i];
		} else {
			System.arraycopy(array, 0, result, 0, length);
		}
		
	}
	
	@Override
	public void set(Object array, int index, Object value)
			throws BindingException {

		if (array instanceof int[]) {
			int[] a = (int[]) array;
			a[index] = (Integer) value;
		} else if (array instanceof boolean[]) {
			boolean[] a = (boolean[]) array;
			a[index] = (Boolean) value;
		} else if (array instanceof byte[]) {
			byte[] a = (byte[]) array;
			a[index] = (Byte) value;
		} else if (array instanceof long[]) {
			long[] a = (long[]) array;
			a[index] = (Long) value;
		} else if (array instanceof float[]) {
			float[] a = (float[]) array;
			a[index] = (Float) value;
		} else if (array instanceof double[]) {
			double[] a = (double[]) array;
			a[index] = (Double) value;
		} else {
			((Object[])array)[index] = value;
		}
		
	}
	
	@Override
	public int size(Object array) throws BindingException {
		if (!isInstance(array)) {
			if(array == null)
				throw new BindingException("Tried to calculate the size of a null array.");
			throw new BindingException("Unexpected class "+array.getClass().getSimpleName()+", "+clazz.getSimpleName()+" expected");
		}
		return Array.getLength(array);
	}
	@Override
	public boolean isInstance(Object obj) {
		return clazz.isInstance(obj);
	}				   

	@Override
	public boolean isImmutable() {
		return true;
	}
	
	@Override
	public void add(Object array, int index, Object element)
			throws BindingException, IndexOutOfBoundsException {
		throw new UnsupportedOperationException();
	}
	
	@Override
	public void remove(Object array, int index, int count) 
	throws BindingException, IndexOutOfBoundsException {
		throw new UnsupportedOperationException();
	}
	
}
