/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.util.binary;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import org.simantics.databoard.util.StreamUtil;

/**
 * Input stream reader
 * 
 * @author Toni Kalajainen (toni.kalajainen@vtt.fi)
 */
public class InputStreamReadable implements BinaryReadable {

	ByteOrder order = ByteOrder.nativeOrder();
	InputStream is;
	long limit, position;
	
	public static BinaryReadable readFully(InputStream is) throws IOException {
		byte[] data = StreamUtil.readFully(is);
		return new BinaryMemory( data );
	}
	
	public InputStreamReadable(InputStream is, long limit)
	{
		this.is = is;
		this.limit = limit;
	}

	/**
	 * Get next byte
	 * @return 0..255
	 * @throws IOException
	 */
	int _get()
	throws IOException
	{
		int value = is.read();
		if (value==-1)
			throw new EOFException();
		position++;
		return value & 0xff;
	}
	
	@Override
	public byte get() 
    throws IOException	
	{
		return (byte) _get();
	}

	@Override
	public void get(byte[] dst, int offset, int length) 
    throws IOException	
	{
		while (length>0) {
			int bytesRead = is.read(dst, offset, length);
			if (bytesRead==-1) throw new EOFException();
			position+=bytesRead;			
			offset += bytesRead;
			length -= bytesRead;
		}
	}

	@Override
	public void get(byte[] dst) 
    throws IOException	
	{
		get(dst, 0, dst.length);
	}

	@Override
	public void get(ByteBuffer buf) 
    throws IOException	
	{		
		for (;buf.hasRemaining();)
			buf.put((byte)_get());		
	}

	@Override
	public void get(ByteBuffer buf, int length) 
    throws IOException	
	{
		if (length<256) {
			for (int i=0; i<length; i++)
				buf.put((byte)_get());
		} else {
			byte[] b = new byte[length];
			get(b, 0, length);
			buf.put(b);
		}
	}

	@Override
	public double getDouble() 
    throws IOException	
	{
		return Double.longBitsToDouble(getLong());
	}

	@Override
	public float getFloat() 
    throws IOException	
	{
		return Float.intBitsToFloat(getInt());
	}

	@Override
	public int getInt() 
    throws IOException	
	{
		if (order == ByteOrder.BIG_ENDIAN)
		{		
			return 
				( _get() << 24) |
				( _get() << 16) | 
				( _get() << 8) |
				( _get() );
		} else {
			return 
				( _get() ) |
				( _get() << 8) |
				( _get() << 16) | 
				( _get() << 24);
		}
	}

	@Override
	public long getLong() 
    throws IOException	
	{
		if (order == ByteOrder.BIG_ENDIAN)
		{		
			return
			( ((long)_get()) << 56) |
			( ((long)_get()) << 48) | 
			( ((long)_get()) << 40) |
			( ((long)_get()) << 32) |		
			( ((long)_get()) << 24) |
			( ((long)_get()) << 16) | 
			( ((long)_get()) << 8) |
			( ((long)_get()) );		
		} else {		
			return 
			( ((long)_get())) |
			( ((long)_get()) << 8) |
			( ((long)_get()) << 16) | 
			( ((long)_get()) << 24) |
			( ((long)_get()) << 32) |		
			( ((long)_get()) << 40) |
			( ((long)_get()) << 48) | 
			( ((long)_get()) << 56);
		}
	}

	@Override
	public short getShort() 
    throws IOException	
	{
		if (order == ByteOrder.BIG_ENDIAN)
		{
			return (short) ( (_get() << 8) |  _get() ) ;
		} else {
			return (short) ( _get() | (_get() << 8) ) ;
		}
	}

	@Override
	public long length() 
	{
		return limit;
	}
	
	@Override
	public long position() {
		return position;
	}
	
	public ByteOrder order() {
		return order;
	}

	public void order(ByteOrder order) {
		this.order = order;
	}

	@Override
	public void skip(long bytes) throws IOException {
		is.skip(bytes);	
	}
	
}

